@@ -12,7 +12,7 @@
 #' Read data from a DBI connection into an ffdf.
 #' 
 #' Read data from a DBI connection into an \code{\link[ff]{ffdf}}. This can for example be used to import
-#' large datasets from Oracle, SQLite, MySQL, PostgreSQL or other SQL databases into R. \cr
+#' large datasets from Oracle, SQLite, MySQL, PostgreSQL, Hive or other SQL databases into R. \cr
 #' 
 #' Opens up the DBI connection using \code{DBI::dbConnect}, sends the query using \code{DBI::dbSendQuery} and \code{DBI::fetch}-es 
 #' the results in batches of next.rows rows. Heavily borrowed from \code{\link[ff]{read.table.ffdf}} 
@@ -340,7 +340,7 @@ read.dbi.ffdf <- function(
 #' Read data from a ODBC connection into an ffdf.
 #' 
 #' Read data from a ODBC connection into an \code{\link[ff]{ffdf}}. This can for example be used to import
-#' large datasets from Oracle, SQLite, MySQL, PostgreSQL or other SQL databases into R. \cr
+#' large datasets from Oracle, SQLite, MySQL, PostgreSQL, Hive or other SQL databases into R. \cr
 #' 
 #' Opens up the ODBC connection using \code{RODBC::odbcConnect}, sends the query using \code{RODBC::odbcQuery} and retrieves
 #' the results in batches of next.rows rows using \code{RODBC::sqlGetResults}. Heavily borrowed from \code{\link[ff]{read.table.ffdf}}
@@ -616,6 +616,291 @@ read.odbc.ffdf <- function(
   return(x)
 }
 
+#' Read data from a JDBC connection into an ffdf.
+#' 
+#' Read data from a JDBC connection into an \code{\link[ff]{ffdf}}. This can for example be used to import
+#' large datasets from Oracle, SQLite, MySQL, PostgreSQL, Hive or other SQL databases into R. \cr
+#' 
+#' Opens up the JDBC connection using \code{RJDBC::dbConnect}, sends the query using \code{RJDBC::dbSendQuery} and \code{RJDBC::fetch}-es 
+#' the results in batches of next.rows rows. Heavily borrowed from \code{\link[ff]{read.table.ffdf}} 
+#'
+#' @param query the SQL query to execute on the JDBC connection
+#' @param dbConnect.args a list of arguments to pass to JDBC's \code{RJDBC::dbConnect} (like drv, dbname, username, password). See the examples.
+#' @param dbSendQuery.args a list containing database-specific parameters which will be passed to to pass to \code{RJDBC::dbSendQuery}.
+#' Defaults to an empty list.  
+#' @param fetch.args a list containing optional database-specific parameters which will be passed to to pass to \code{RJDBC::fetch}. 
+#' Defaults to an empty list.          
+#' @param x NULL or an optional ffdf object to which the read records are appended. 
+#' See documentation in read.table.ffdf for more details and the example below.
+#' @param nrows Number of rows to read from the query resultset. Default value of -1 reads in all rows.
+#' @param first.rows chunk size (rows) to read for first chunk from the query resultset
+#' @param next.rows chunk size (rows) to read sequentially for subsequent chunks from the query resultset. Currently, this must be specified.
+#' @param levels optional specification of factor levels. A list with as names the names the columns of the data.frame 
+#' fetched in the first.rows, containing levels of the factors. 
+#' @param appendLevels logical. A vector of permissions to expand levels for factor columns. See documentation in \code{\link[ff]{read.table.ffdf}} for more details.
+#' @param asffdf_args further arguments passed to \code{\link[ff]{as.ffdf}} (ignored if 'x' gives an ffdf object )
+#' @param BATCHBYTES integer: bytes allowed for the size of the data.frame storing the result of reading one chunk. 
+#' See documentation in \code{\link[ff]{read.table.ffdf}} for more details.
+#' @param VERBOSE logical: TRUE to verbose timings for each processed chunk (default FALSE).
+#' @param colClasses See documentation in \code{\link[ff]{read.table.ffdf}}
+#' @param transFUN function applied to the data frame after each chunk is retreived by \code{RJDBC::fetch}
+#' @param ... optional parameters passed on to transFUN
+#' @return 
+#' An ffdf object unless the query returns zero records in which case the function will return the data.frame
+#' returned by \code{RJDBC::fetch} and possibly transFUN. 
+#' @export
+#' @seealso \code{\link[ff]{read.table.ffdf}, \link{read.jdbc.ffdf}}
+#' @author Jan Wijffels
+#' @examples
+#' \dontrun{
+#' require(ff)
+#' 
+#' ##
+#' ## Example query using data in sqlite
+#' ##
+#' require(RSQLite)
+#' dbfile <- system.file("smalldb.sqlite3", package="ETLUtils")
+#' drv <- JDBC(driverClass = "org.sqlite.JDBC", classPath = "/usr/local/lib/sqlite-jdbc-3.7.2.jar")
+#' query <- "select * from testdata limit 10000"
+#' x <- read.jdbc.ffdf(query = query, dbConnect.args = list(drv = drv, url = sprintf("jdbc:sqlite:%s", dbfile)), 
+#'                     first.rows = 100, next.rows = 1000, VERBOSE=TRUE)
+#' class(x)
+#' x[1:10, ]
+#' }
+read.jdbc.ffdf <- function(
+  query = NULL,
+  dbConnect.args = list(drv=NULL, dbname = NULL, username = "", password = ""), 
+  dbSendQuery.args = list(), 
+  fetch.args = list(),   	
+  x = NULL, nrows = -1, 
+  first.rows = NULL, next.rows = NULL, levels = NULL, appendLevels = TRUE, 
+  asffdf_args = list(), BATCHBYTES = getOption("ffbatchbytes"), VERBOSE = FALSE, colClasses = NULL, 
+  transFUN = NULL, ...){
+  require(RJDBC)
+  cleanupConnection <- function(x){
+    if("resultset" %in% names(x)){
+      dbClearResult(res=x$resultset)
+    }
+    if("channel" %in% names(x)){
+      dbDisconnect(x$channel)
+    }
+  }
+  jdbcinfo <- list()
+  
+  ##
+  ## Connect to database
+  ##
+  jdbcinfo$channel <- do.call('dbConnect', dbConnect.args)
+  on.exit(cleanupConnection(jdbcinfo))
+  
+  append <- !is.null(x)
+  if(append && !inherits(x, "ffdf")){
+    stop("only ffdf objects can be used for appending (and skipping the first.row chunk)")
+  }
+  if(VERBOSE){
+    read.time <- 0
+    write.time <- 0
+  }
+  
+  nrows <- as.integer(nrows)
+  N <- 0L	
+  if(!append){		
+    if(VERBOSE){
+      cat("read.jdbc.ffdf ", N+1L, "..", sep="")
+      read.start <- proc.time()[3]
+    }		
+    if(is.null(colClasses)){
+      colClasses <- NA
+    }		
+    ## Specify the size of the first number of rows to import
+    if (is.null(first.rows)){
+      if (is.null(next.rows)){
+        first.rows <- 1e3L
+      }else{
+        first.rows <- as.integer(next.rows)
+      }				
+    }else{
+      first.rows <- as.integer(first.rows)
+    }		
+    if (nrows>=0L && nrows<first.rows){
+      first.rows <- nrows
+    }		
+    if (first.rows==1){
+      stop("first.rows must not be 1")
+    }
+    ##
+    ## Set up arguments for odbcQuery and sqlGetResults functions
+    ##
+    dbSendQuery.args$conn <- jdbcinfo$channel
+    dbSendQuery.args$statement <- query
+    ##
+    ## Launch the query
+    ##
+    jdbcinfo$resultset <- do.call('dbSendQuery', dbSendQuery.args)
+    
+    ##
+    ## Read in the first chunk
+    ##
+    fetch.args$res <- jdbcinfo$resultset
+    fetch.args$n <- first.rows			
+    if(is.null(transFUN)){
+      dat <- convertCharacterToFactor(do.call("fetch", fetch.args))
+    }else{
+      dat <- convertCharacterToFactor(transFUN(do.call("fetch", fetch.args), ...))
+    }				
+    n <- nrow(dat)
+    N <- n
+    ## Recode levels if levels are specified
+    if(!is.null(levels)){
+      cnam <- colnames(dat)
+      lnam <- names(levels)
+      if(!is.list(levels) || is.null(lnam) || any(is.na(match(lnam,cnam)))){
+        stop("levels must be a list with names matching column names of the first data.frame read")
+      }				
+      for (i in lnam){
+        dat[[i]] <- recodeLevels(dat[[i]], levels[[i]])
+      }
+    }		
+    if (VERBOSE){
+      write.start <- proc.time()[3]
+      read.time <- read.time + (write.start - read.start)
+      cat(N, " (", n, ")  jdbc-read=", round(write.start-read.start, 3), "sec", sep="")
+    }
+    ##
+    ## If there are zero rows in the query, return a data.frame instead of an ffdf object
+    ##
+    if(nrow(dat) == 0){
+      if (VERBOSE){
+        cat(" query returned 0 records\n", sep="")
+        cat(" jdbc-read=", round(read.time, 3), "sec  TOTAL=", round(read.time, 3), "sec\n", sep="")
+      }				
+      return(dat)
+    }
+    
+    ##
+    ## If there are non-zero rows in the query, convert to a filebased ffdf object
+    ##
+    x <- do.call("as.ffdf", c(list(dat), asffdf_args))		
+    ##
+    ## Fix ordered factors
+    ##
+    colClasses <- repnam(colClasses, colnames(x), default=NA)
+    i.fix <- seq.int(length.out=ncol(dat))[!is.na(match(colClasses, "ordered"))]
+    for (i in i.fix){
+      virtual(x[[i]])$ramclass <- c("ordered","factor")
+    }		
+    if (VERBOSE){
+      write.stop <- proc.time()[3]
+      write.time <- write.time + (write.stop - write.start)
+      cat(" ffdf-write=", round(write.stop-write.start, 3), "sec\n", sep="")
+    }
+  }
+  ##
+  ## Fetch the other rows
+  ##
+  if (append || N==first.rows){
+    ##
+    ## Define the number of rows to fetch in each run depending on the BATCHBYTES maximum
+    ##
+    k <- ncol(x)
+    col.names <- colnames(x)
+    colClasses <- sapply(seq.int(length.out=ncol(x)), function(i)colClass(x[[i]]))		
+    if (is.null(next.rows)){
+      recordsize <- sum(.rambytes[vmode(x)])
+      next.rows <- BATCHBYTES %/% recordsize
+      if (next.rows<1L){
+        next.rows <- 1L
+        warning("single record does not fit into BATCHBYTES")
+      }
+    }else{
+      next.rows <- as.integer(next.rows)
+    }		
+    next.nrows <- next.rows
+    
+    appendLevels <- repnam(appendLevels, col.names, default=TRUE)
+    if(any(appendLevels)){
+      i.fac <- seq.int(length.out=k)
+      if(append == TRUE){
+        i.fac <- i.fac[appendLevels & sapply(i.fac, function(i) is.factor(x[[i]]))]
+      }else{
+        i.fac <- i.fac[appendLevels & sapply(i.fac, function(i) is.factor(dat[[i]]))]	
+      }
+      
+    }	
+    if(append == TRUE){
+      ## Set up arguments for odbcQuery and sqlGetResults functions & Launch the query
+      dbSendQuery.args$conn <- jdbcinfo$channel
+      dbSendQuery.args$statement <- query			
+      jdbcinfo$resultset <- do.call('dbSendQuery', dbSendQuery.args)
+    }
+    ##
+    ## Loop - fetching the subsequent data in batches
+    ##
+    while(TRUE){
+      if (nrows>=0L && N+next.rows > nrows){
+        next.nrows <- nrows - N
+      }				
+      if (next.nrows<1L){
+        break
+      }			
+      if (VERBOSE){
+        cat("read.jdbc.ffdf ", N+1L,"..", sep="")
+        read.start <- proc.time()[3]
+      }
+      ##
+      ## read in subsequent chunk
+      ##
+      fetch.args$res <- jdbcinfo$resultset
+      fetch.args$n <- next.nrows
+      if (is.null(transFUN)){
+        dat <- convertCharacterToFactor(do.call("fetch", fetch.args))
+      }else{
+        dat <- convertCharacterToFactor(transFUN(do.call("fetch", fetch.args), ...))
+      }			
+      n <- nrow(dat)
+      N <- N + n
+      if (VERBOSE){
+        write.start <- proc.time()[3]
+        read.time <- read.time + (write.start - read.start)
+        cat(N, " (", n, ")  jdbc-read=", round(write.start-read.start, 3), "sec", sep="")
+      }
+      ## Nothing to import any more
+      if (n<1L){
+        if (VERBOSE)
+          cat("\n")
+        break
+      }
+      ## Recode levels if levels are specified
+      if(any(appendLevels))
+        for (i in i.fac){
+          lev <- unique(c(levels(x[[i]]),levels(dat[[i]])))  # we save a call to the more general appendLevels() here
+          levels(x[[i]]) <- lev
+          dat[[i]] <- recodeLevels(dat[[i]], lev)
+        }
+      ##
+      ## Append to the ffdf object
+      ##
+      nff <- nrow(x)
+      nrow(x) <- nff + n
+      i <- hi(nff+1L, nff+n)
+      x[i,] <- dat
+      
+      if (VERBOSE){
+        write.stop <- proc.time()[3]
+        write.time <- write.time + (write.stop - write.start)
+        cat(" ffdf-write=", round(write.stop-write.start, 3), "sec\n", sep="")
+      }			
+      if(n<next.nrows){				
+        break
+      }				
+    }		
+  }
+  if (VERBOSE){
+    cat(" jdbc-read=", round(read.time, 3), "sec  ffdf-write=", round(write.time, 3), "sec  TOTAL=", round(read.time+write.time, 3), "sec\n", sep="")
+  }	
+  return(x)
+}
+
 
 
 
